#version 330
out float frag_color;

in vec2 p1;
in vec2 p2;
in vec2 p;
in float depth;

uniform float stroke_width;
uniform vec2 resolution;
uniform bool is_vanishing_mode;
uniform vec2 output_vanishing_point;
uniform int ssaa;
uniform ivec2 face_offset;
uniform sampler2D depth_tex;

float sd_segment(vec2 p, vec2 a, vec2 b, out float t) {
    // Return the distance from a point `p` to a line segment defined by its end
    // points `a` and `b` and put the t-parameter for the closest point on the
    // line segment to `p` into `t`
    vec2 ap = p - a;
    vec2 ab = b - a;
    t = clamp(dot(ap, ab) / dot(ab, ab), 0.0, 1.0);
    return distance(ap, ab * t);
}

float sdfOutline(float sd, float width);

void main() {
    float t;
    frag_color = sdfOutline(sd_segment(p, p1, p2, t), stroke_width);
    vec2 sample_point = p;
    
    if (
        is_vanishing_mode &&
        p.x < output_vanishing_point.x &&
        abs(p.y - output_vanishing_point.y) < stroke_width * 0.5 + 1.0
    ) {
        sample_point = mix(p1, p2, t);
        sample_point.x = clamp(sample_point.x, 0.5, resolution.x - 0.5);

        if (abs(max(p1, p2).y - output_vanishing_point.y) >= abs(min(p1, p2).y - output_vanishing_point.y)) {
            float bound = round(output_vanishing_point.y + 0.1);
            sample_point.y = max(sample_point.y, bound + 0.5);
        } else {
            float bound = round(output_vanishing_point.y - 0.1);
            sample_point.y = min(sample_point.y, bound - 0.5);
        }
    }
    
    if (depth > texelFetch(depth_tex, face_offset + ivec2(sample_point * float(ssaa)), 0).r) {
        discard;
    }
}

// The following functions are from https://www.shadertoy.com/view/mtcXDH

float areaSquareLine(vec2 n, float d) {
    // https://www.desmos.com/calculator/dorvdj5nbq for visualization
    n = abs(n);
    n = vec2(max(n.x, n.y), min(n.x, n.y));
    float a = abs(d);
    float b;

    if (n.y != 0.0 && n.x - n.y <= 2.0 * a) {
        vec2 c = 1.0 + (n - 2.0 * a) / n.yx;
        b = 1.0 - max(c.x, 0.0) * c.y / 8.0;
    } else {
        b = min(0.5 + a / n.x, 1.0);
    }
    
    return d < 0.0 ? 1.0 - b : b;
}

float areaSquareLineAverage(float d) {
    // Approximate average value of areaSquareLine(n, d) w.r.t. n
    float a = abs(d);
    float b;
    
    if (a < 1.0 / sqrt(2.0)) {
        b = 4.1434218 * pow(a, 12.647891) -
            1.3070394 * pow(a, 3.9787831) +
            1.0998631 * a +
            0.5012205;
        b = clamp(b, 0.0, 1.0);
    } else {
        b = 1.0;
    }
    
    return d < 0.0 ? 1.0 - b : b;
}

float sdfFill(float sd) {
    // Return the fraction of the pixel covered by the SDF (i.e. where
    // sd < 0.0)
    vec2 n = vec2(dFdx(sd), dFdy(sd));
    
    if (n == vec2(0.0)) {
        // Ambiguous case, so instead we just return the average. This
        // is the problem with using dFdx/dFdy that would occur much
        // less frequently if we had access to the analytic derivatives.
        return areaSquareLineAverage(-sd);
    }

    n = normalize(n);
    return areaSquareLine(n, -sd);
}

float sdfOutline(float sd, float width) {
    // Return the fraction of the pixel covered by the outline of the
    // SDF (i.e. where abs(sd) < width / 2.0)
    return sdfFill(sd - width / 2.0) - sdfFill(sd + width / 2.0);
}
