#version 330
layout (location = 0) in vec3 a_p1;
layout (location = 1) in vec3 a_p2;
layout (location = 2) in float a_z;

out vec2 p1;
out vec2 p2;
out vec2 p;
out float depth;

uniform mat4 mvp;
uniform vec2 resolution;
uniform float stroke_width;
uniform int extrusion_mode;
const int DIRECTIONAL = 1;
const int VANISHING = 2;
const int CAMERA = 3;
uniform vec2 extrusion_direction;
uniform vec2 extrusion_vanishing_point;
uniform float back_extrusion_amount;
uniform float front_extrusion_amount;
uniform mat4 path_to_screen;

vec4 extrude(vec3 p) {
    // Extrude the path-space position `p` into it's clip-space position
    // p.z ==  0.0: front face
    // p.z == -1.0: back face
    float extrusion_amount = mix(front_extrusion_amount, back_extrusion_amount, -p.z);
    
    switch (extrusion_mode) {
        case DIRECTIONAL:
            return vec4(p.xy + extrusion_direction * extrusion_amount, p.z, 1.0);
        case VANISHING:
            return vec4(mix(p.xy, extrusion_vanishing_point, extrusion_amount), p.z, 1.0);
        case CAMERA: {
            return path_to_screen * vec4(p.xy, -extrusion_amount, 1.0);
        }
    }
}

vec2 world_to_screen(vec2 p) {
    return ((mvp * vec4(p, 0.0, 1.0)).xy * 0.5 + 0.5) * resolution;
}

vec4 screen_to_clip(vec3 p) {
    return vec4(p.xy / resolution * 2.0 - 1.0, -2.0 * p.z - 1.0, 1.0);
}

void main() {
    vec4 c1 = extrude(a_p1);
    vec4 c2 = extrude(a_p2);
    p1 = c1.xy / c1.w;
    p2 = c2.xy / c2.w;
    float zw1 = c1.z + c1.w;
    float zw2 = c2.z + c2.w;
    float ti = zw1 / (zw1 - zw2);
    vec4 c = mix(c1, c2, clamp(ti, 0.0, 1.0));
    vec2 w = c.xy / c.w;
    gl_Position = vec4(0.0);
    
    if (zw1 < zw2) {
        if (ti >= 1.0) {
            return;
        }
        
        p1 = w;
    } else if (zw1 > zw2) {
        if (ti <= 0.0) {
            return;
        }

        p2 = w;
    } else if (zw1 < 0.0) {
        return;
    }
    
    p1 = world_to_screen(p1);
    p2 = world_to_screen(p2);

    vec2 t = normalize(p2.xy - p1.xy);
    t *= 0.5 * (stroke_width + abs(t.x) + abs(t.y));
    vec2 n = vec2(-t.y, t.x);
    int i = gl_VertexID % 4;
    p = (i < 2 ? p1.xy - t : p2.xy + t) + (i % 2 == 0 ? -n : n);
    gl_Position = screen_to_clip(vec3(p, a_z));
    depth = -a_z;
}
